// ================================================================
//  Educational Game Platform - with a top level Menu selection
//  using a Micro Controller, 8 LED 4 color bar and 4 Button module
//

//  include Library of Hardware Support functions
#include "Mini_STEM_Platform.h"

unsigned long t0,t1;

// the setup routine runs once on power up or when you press reset:
//=================================================================
void setup() {
  int aVal, val;
  
  // initialize serial communication at 115200 bits per second:
  Serial.begin(115200);
  Serial.println("Arduino Educational Mini Game Platform - Lesson 11");

  // ------------ setup the Mini Game/STEM platform (8LEDs, 4Btns, and Audio)
  gPlatform_setup();

  beep(50,50); beep(50,50); beep(150);  // announce "it is alive!"

  // wait an amount of pseudo random time due to HW input noise (Entropy), then randomize
  aVal = val = analogRead(A0);
  while (abs(aVal-val)<=20) {val=analogRead(A0); delayMicroseconds(micros()%val);}    // expects analog noise (A0 needs to be open)
  randomSeed(micros()/4);  // micros() resolution is ~ 4us
}

// the loop routine runs over and over again forever:
//=================================================================
void loop() { // Top Menu level, for selecting 1 of N activities
  byte func, maxFunc;

  func=1;
  while(1) {
    clearDisp();
    while(Btn) {buZZ(1); scanBtns();}  // wait for no button

    maxFunc=6;  // currently there are 6 activities/games supported
    // indicate selectable functions 1-6
    dim[1]=dim[2]=dim[3]=dim[4]=dim[5]=dim[6]=true;
    brightOne=func;
    refreshDisp();
    
//    while(Btn) {scanBtns(); refreshDisp();}  // wait here for button release

    while (! btn1) {
      refreshDisp();
      delay(1);
      scanBtns();
      if (btn2Pressed) brightOne = (brightOne==1)?  maxFunc : brightOne-1;  // -- move left
      if (btn3Pressed) brightOne = (brightOne%maxFunc) + 1; // ++ moves right
      if (btn4Pressed) brightOne = (brightOne%maxFunc) + 1; // allow most play with only 2 buttons (1&4)
    }
////    randomSeed(micros());  // Add "more randomness" to random() calls, based on time btn pressed

    func = brightOne;
    clearDisp();
    while(Btn) scanBtns(); // wait for button release

    Serial.print(" FUNC: "); Serial.println(func);
    switch (func) {
      case 1:
        HW_Support_Demo();  // Demonstrates LED display and audio support functionality
                           // New & Improved intro demonstration
        break;
      case 2:
        Speed_Scan_ver4();   // hit it on the right spot to speed things up to a winning max
        break;
      case 3:
        React_Now_ver2();    // Test your Reaction time
        break;
      case 4:
        SeeIt_TagIt();      // Capture and Release fleeting blips
        // btnReporting();    // exercises the buttons and Leds, and reports via Serial
        // Or replace this with your own previously devised activity
        break;
      case 5:
        pingPong();     // Two person game of Ping-Pong
        break;
      case 6:
        Magic_8_ball_v3(); // Give responses using PRNG along with Human Heuristics
        break;
      default:
        buZZ(1000);
    }

  }
}

// Various selection of mini APPs / Games
//=================================================================

// ------------------
void btnReporting() {
  // =================================  Button - LED exercise & Reporting
  int icnt=0; // iteration count, lsb~=1msec
  clearDisp();
  while (1) {
    refreshWait(1);
    icnt++;
    scanBtns();
    // ------------------- button reporting
    if (BtnChanged) {
      Serial.println(" ");
      if (btn1) Serial.print("[Btn1] "); else Serial.print("       ");
      if (btn2) Serial.print("[Btn2] "); else Serial.print("       ");
      if (btn3) Serial.print("[Btn3] "); else Serial.print("       ");
      if (btn4) Serial.print("[Btn4] "); else Serial.print("       ");
      //Serial.print(btnState);

      // ------------------ LED reflection of button state
      lit[1] = lit[2] = btn1; // set Red LEDs
      lit[3] = lit[4] = btn2; // set Yel LEDs
      lit[5] = lit[6] = btn3; // set Blu LEDs
      lit[7] = lit[8] = btn4; // set Grn LEDs
      brightOne=0;

      beep(15); // make a key press click, for up&down
      icnt=0;
    } else if (icnt>10000) {  // if ignorned for 10 seconds
      icnt=0;
      brightOne=1;
      beep(15,200);
      clearDisp();
    }
    if (ESC && brightOne==0 && btn4 ) Serial.print(" [ESC] ");    
    brightOne = (ESC)? 8 : 0;
    if (ESC && btn1) break;    // only leave Btn Reporting if btn1 is also held.
  }
}

void HW_Support_Demo() {  // Demonstrates use of set of Hardware support functions
  bool WelcomeScreen=true;

  //  demostrate the usefulness of set_LEDs()
  for (int i=0;i<8;i++) { refreshWait(100); set_LEDs(1<<i,1); }
  for (int i=0;i<8;i++) { refreshWait(100); set_LEDs(1<<i,2); }
  for (int i=0;i<8;i++) { refreshWait(100); set_LEDs(1<<i,3); }
  for (int i=0;i<8;i++) { refreshWait(100); set_LEDs(1<<i,0); }
  set_LEDs(0xAA,1); // set dim-BRT led pattern
  set_LEDs(0x55,3);
  refreshWait(800);
  clearDisp();
  
  while(true) {
    scanBtns();

    if (WelcomeScreen && !BtnPressed) {  // display splash display until a new button presses
      for (int i=1; i<=nleds; i++) {
        dim[i]=true;    // set all LEDs dim
        lit[i] = ((msCnt/333+i)%2);   // flash odd & even LEDs in opposition
      }
//      refreshWait(1);
    } else WelcomeScreen=false;

    if (BtnPressed) {
      clearDisp();
      for (int i=1; i<=nleds; i++) {  // set all LEDs to a given level
        if (btn1Pressed) dim[i]=true;
        if (btn2Pressed) lit[i]=true;
        if (btn3Pressed) brt[i]=true;
        //if (btn4Pressed) WelcomeScreen=true;
      }
    }

    // produce sound while btn1-3 is held
    while (Btn) {
      if (btn1 || btn3) boop(2);
      if (btn2 || btn3) beep(2);
      if (btn4) break;  // don't get stuck here
      scanBtns();
    }
    
    refreshDisp();
    delay(1);
//
    if (ESC) break;
  } 
}

void Speed_Scan_ver3() {
//  int btnState, lastState;
//  bool btnPressed;
  int onMSecs, stRate=165;
  int blip, hits=0, goal=10;

  blip=1;
  onMSecs=stRate;
  while(true) { 
    //digitalWrite(lites[blip], ON_STATE);
    lit[blip]=true;
    refreshWait(onMSecs);     // flash the LED (blip) for a bit

    // allow use of left or right most button
    scanBtns();
//    btnState=(digitalRead(button[0])==PRESSED_STATE || digitalRead(button[3])==PRESSED_STATE);
//    btnPressed = btnState!=lastState && btnState;
//    lastState=btnState;

    // note blip 4||5 when to 5||6
    if (BtnPressed && (blip==5 || blip==6)) {
        beep(100);    
        hits++;
        if (hits==goal) {
          beep(100,50); beep(100,50); beep(250);  // you did it
          hits=0; // start over
          refreshWait(1000);
        }
    }

    // added block for support of a missed penalty
    // note blip 4||5 when to 5||6
    if (BtnPressed && !(blip==5 || blip==6)) {
      refreshWait(300);
      hits--;     // slow down update rate
    }
    
    onMSecs = stRate - (15*hits); // more hits he get faster the blip scans
    //digitalWrite(lites[blip], !ON_STATE);  // turn last lit LED off
    lit[blip]=false;
    //blip = (blip+1) % nleds;    // advance to next LED
    // next line changed to cycle from 1 to nleds
    blip = (blip % nleds) + 1;    // advance to next LED
  }
}

void Speed_Scan_ver4() {  // hit it on the right spot to speed things up to a winning max
                          // misses slows things down, and can put you even further from winning
  int onMSecs, stRate=150;
//  int blip, hits=0, goal=10;
  int hits=0, goal=9;

  litLED=1;
  onMSecs=stRate;
  while(true) { 
//    lit[blip]=true;
    refreshWait(onMSecs);     // flash the LED (blip) for a bit

    // allow use of left or right most button
    scanBtns();

    if (BtnPressed && (litLED==5 || litLED==6)) {
        beep(100);    
        hits++;
        if (hits==goal) {
          beep(100,50); beep(100,50); beep(250);  // you did it
          hits=0; // start over
          refreshWait(1000);
          clearDisp();
        }
    }

    // added block for support of a missed penalty
    if (BtnPressed && !(litLED==5 || litLED==6)) {
      refreshWait(300);
      hits--;     // slow down update rate
      clearDisp();
    }
    
    onMSecs = stRate - (15*hits); // more hits he get faster the blip scans
//    lit[blip]=false;
    for (int i=1; i<=hits; i++) dim[i]=true;    // display as, dim, the hit progress
//    blip = (blip % nleds) + 1;    // advance to next LED
    litLED = (litLED % nleds) + 1;    // advance to next LED
//
    if (ESC) break;
  }
}

//----------------  React_Now - Swat Quickly as LEDs light up
void React_Now() { // attemp to hit the button when the first LED lights up
                   // 1st LED is lit for 100 msecs, additional LEDs lite every 20ms
                   // challenge: if one pushes & holds before the first LED they get a wrong 'win'; how to fix ?
  bool btn = !PRESSED_STATE;
  int min_Period = 100;
  int btnInput;
  int LEDptr;
  
  btnInput = button[0];   // use the first button on the left of the four
  while(true) {
    
    // insure that the button is not already pressed
    while (btn==PRESSED_STATE) btn=digitalRead(btnInput);

    // insure all LEDs are off
    for (int LEDptr=0; LEDptr<8; LEDptr++) digitalWrite(lites[LEDptr], !ON_STATE);

    delay(random(2000,5000));       // wait for 2 - 5 seconds
    digitalWrite(lites[0], ON_STATE);
    beep(10); // takes 2x msecs
    delay(min_Period-20-20);    // account for the 20ms beep time and 20ms in for loop

    for (LEDptr=0; LEDptr<8; LEDptr++) {
      digitalWrite(lites[LEDptr], ON_STATE);
      delay(20);
      btn = digitalRead(btnInput);
      if (btn==PRESSED_STATE) { // they pressed the button
        delay(1000);
        break;
      }
    }
    if (LEDptr==8) beep(100);
  }
}

// --------------- See It ... Hit It !    (aka Capture & Release)
// Try capturing all 4 fleeting blips before releasing them one at a time, without fault
void SeeIt_TagIt() {    // attempt to hit assosiated button when an LED lights up
  int time_given = 330; // within given time
  byte n;
  
  int ver=1;  // =1 for lenient, =2 for strickly the right & only the right btn pressed
  clearDisp();  // insure all LEDs are off
  while(true) {
    refreshWait(random(1000,4000));       // wait for 1 - 4 seconds
    //n = random(4);    // pick an LED
    n = (random(1,4) + n) % 4;    // pick an LED, other than the last one picked 
    brightOne=2*n+1;
    refreshWait(time_given);
    scanBtns();
    if (btn[n+1] && !(ver==2 && (btns || !BtnChanged))) {
      lit[2*n+1]=!lit[2*n+1];
      beep(100,100); boop(400);  // ta Da!
      while(Btn) scanBtns();  // stay until released
    }
    brightOne=0;
    
    scanBtns();
    if (ESC) break;
  }
}

//----------------  React_Now - Swat Quickly as LEDs light up
void React_Now_ver2() { // attemp to hit the button when the first LED lights up
                   // 1st LED is lit for 100 msecs, additional LEDs lite every 20ms
                   // challenge: if one pushes & holds before the first LED they get a wrong 'win'; how to fix ?
//  bool btn = !PRESSED_STATE;
  int min_Period = 100;
//  int btnInput;
  int LEDptr;
  
//  btnInput = button[0];   // use the first button on the left of the four
  while(true) {
//    scanBtns();
    // insure that the button is not already pressed
    while (Btn && !ESC) scanBtns();
    if (ESC) return;
    
    // insure all LEDs are off
    clearDisp();

    delay(random(2000,5000));       // wait for 2 - 5 seconds
    lit[1]=1;
    refreshDisp();
    beep(10); // takes 2x msecs
    delay(min_Period-20-20);    // account for the 20ms beep time and 20ms in for loop

    for (LEDptr=1; LEDptr<=8; LEDptr++) {
//      digitalWrite(lites[litLED], ON_STATE);
//      delay(20);
      lit[LEDptr]=true;
      refreshWait(20);
      scanBtns();
      if (BtnPressed) { // they pressed the button
        brightOne=LEDptr;
        refreshWait(1000);
        break;
      }
    }
    if (LEDptr>8) beep(100);
//    scanBtns();
    if (ESC) break;
  }
}

// =====================================================================
void pingPong() { // Ping-Pong - Two player Game
  #define STARTPERIOD 330;
  int litPeriod, period;
  int cnt=0, maxCnt;
  int rally=0;
  int dir = 1;
  byte color; // 1 for left side, 4 for right
  bool isLit=false;

  bool lastSwung, swing=false;
  bool lastPosHit, hit=false;
  bool missed=false;
  bool badSwing=false;
  int ver=1;
                      // this version: for right:Btn1|Btn / left:btn4|Btn0
  litPeriod = STARTPERIOD;
  period = litPeriod;
  maxCnt = nleds-1;
  t0 = millis();

while(1) {
  lastSwung = swing;
//  swing =  btnPressed();
  scanBtns();
  if (ESC) return;
  if (dir==1) swing = btn4;
  else        swing = btn1;
  hit = hit || swing;
  if (btn2Pressed) ver=2;  // repeat until they swing at it 
  if (btn3Pressed) ver=3;  // give endless attemps

  if ((millis() - t0) >= period) {  // is this time period up ?
    if (isLit) {                       // about to go from LED on to OFF
//      digitalWrite(ONBOARD_LED, LOW);
      period = litPeriod / 3;
      if (1) {
        missed=false;
        if (cnt>=maxCnt || cnt<=0 || badSwing) {  // ------ end of the line
//            if (cnt<=0) dir = 1; else  dir = -1;
//            Serial.print("rally = "); Serial.println(rally);
            //litPeriod = 9 * litPeriod / 10;
            litPeriod = STARTPERIOD;
            for (int i = 1; i < rally; i++) {
                litPeriod = 9 * litPeriod / 10;   // * .9
            }
//            Serial.print("litPeriod = "); Serial.println(litPeriod);

//        !!!!!!!!!!!!!!!!  make judgement
          if (hit && !lastPosHit) {
            // he hit it
            litLED = cnt+1;
            refreshWait(40);
            litLED = 0;
            hit=false;
            rally++;  
            if (cnt<=0) dir = 1; else  dir = -1;
          } else {  // failed to hit it
              litLED=0;
              clearDisp();
              color = (dir>0)? 1: 4; 
              // acknowledge the "Win-ner!" 
              if (rally<1 && !hit) {
                if (color==1) {color=1;}
                else          {color=4;}
                missed=true;
              } else if (ver==3 || (ver==2 && !hit && !badSwing)) {
                cnt = (dir==1)? 0 : nleds-1;
              } else {
                if (color==1) {brt[1]=1; beep(150); delay(100); brt[2]=1; beep(500);}
                else          {brt[nleds-1]=1; beep(150); delay(100); brt[nleds]=1; beep(500);}
                missed=true;
              }
              
              if (missed) {
                color=0;
                refreshWait(1000);
                if (rally>=2) {
                  showLevel(rally/2);   // give em one star for each set of back & forth
                  refreshWait(2000);
                  brightOne=0;
                  refreshWait(200);
                }
                // re-init
                litPeriod = STARTPERIOD;
                period = litPeriod;
                rally =0;
                
                // change direction
                if (color==1) {
                  dir = 1;
                  cnt = 0;
                } else {
                  dir = -1;
                  cnt = nleds-1;
                }
              }
              clearDisp();
          }
          badSwing=false;
        } // end time to judge
        else {
          if (hit && ver!=3) {  // on wrong cell
            badSwing=true;
            buZZ(30);
          }
        } 
      }
    } else {                     // about to go from LED off to ON
      period = litPeriod;
      lastPosHit = hit;
      hit = false;
      cnt = cnt+dir;    // goto NEXT cell      
    }
    t0 = millis();
    isLit = not isLit;
//    pressed = false;
  }

  //-----------  UPDATE Display
    for (int i=1; i<=nleds; i++)  dim[i]=0;
    dim[cnt+1] = 1;
    if (color) showColor(color);
    refreshDisp();

    delay(1);
    if (ESC) return;
  }
}


// ==========================================  Magic-8-Ball
// hold down btn1 ... ponder question, let go.  Red: NO,  YEL: Maybe,  Grn: YES  Blue: Try again
// use btn2 for 1/6 out comes ~ like a dice, useful for games.  (use btn3 for 1/8)
void Magic_8_ball_v3() {
  byte ans, keyCode, color=0;
  byte priorBtn;
  static byte dicePat[] = {0x10, 0x42, 0x2A, 0xA5, 0xD3, 0xE7};
  
  while (1) {
    clearDisp();
    while(!Btn) {
      scanBtns();
      litLED=(msCnt/1000)%2;    // toggles 0/1 per second
      refreshWait(1);
    }
    if (ESC || btn4) break;
    
    t0 = millis();  // start timing on  Button pushed
    litLED=0;

    while(Btn) {  // wait for release
      // Mere  potential answers periodically come up during the posing of a question.
      priorBtn=btnNum;
      scanBtns();
      if ((msCnt%333)==1) {
        if ((msCnt/333)%2 == 1) clearDisp();
        else {
          // determine next color set to light or not to light
          if (color>2) color=5-color;
          else color+=2;
          clearDisp();
          showColor(color);
        }
      }  // flash odd & even LEDs in opposition
      refreshWait(1);
    }

    keyCode = priorBtn;
    clearDisp();
    refreshWait(300 + random(800));  // make them wait alittle
    
    color=0;
    if (keyCode==1) {            // select one of the 4 colors
      ans = (millis() - t0) %4;  // answer is determined by their action
      showColor(ans+1);
    } else if (keyCode==2) {     // select one of the 6 colors , can be used as die for for game dice 
      ans = (millis() - t0) %6;  // answer is determined by their action
      //brightOne = ans+1;
      set_LEDs(dicePat[ans],2);   // lite a given pattern, similar to dice, for 1-6
    } else {                     // btn3 was used, 1 of N leds
      ans = (millis() - t0) %nleds;
      brightOne = ans+1;
    }

    while(1) {      // await the next Request
      scanBtns();
      refreshWait(1);
      if (ESC) return;     // ESC when done
      if (Btn) break;      // go again 
    }
  }
}

// ================== Local Support Utility Functions =================

void showColor(byte color) {
  if (color==0) return;
  if (color==1 || color==5)  lit[1] = lit[2] = true; // set Red LEDs
  if (color==2 || color==5)  lit[3] = lit[4] = true; // set Yel LEDs
  if (color==3 || color==5)  lit[5] = lit[6] = true; // set Blu LEDs
  if (color==4 || color==5)  lit[7] = lit[8] = true; // set Grn LEDs
}

// -----------------------
void showLevel(byte level) {
  int n;
  clearDisp();
  refreshWait(250);
  for (int i=0; i<level; i++) {
    beep(10);
    n = (i%nleds)+1;  // n wraps around as i exceeds nleds
    dim[n]=true;
    brightOne=n;
    refreshWait(100);
  }
}

void set_LEDs(byte led_pattern, byte state) {
  // led_pattern: bit decoded pattern, eg: 9= (00001001b)left most led + forth (led#1 & #4)
  // state: 1=dim, 2=lit, 3=bright, 0=clear
  byte b;
  for (int i=1; i<=8; i++) {
    b = led_pattern & 1;
    led_pattern=led_pattern>>1;
    if (b) {
      dim[i]=0; lit[i]=0; brt[i]=0; // default state cleared
      if (state==1)  dim[i] = true;
      else if (state==2)  lit[i] = true;
      else if (state==3)  brt[i] = true;
    }
  }
}
